package ca.recrown.islandsurvivalcraft.world.generation;

import java.util.LinkedList;
import java.util.Random;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;

import org.bukkit.Chunk;
import org.bukkit.Material;
import org.bukkit.World;
import org.bukkit.block.Biome;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.world.ChunkLoadEvent;
import org.bukkit.generator.ChunkGenerator;

import ca.recrown.islandsurvivalcraft.Utilities;
import ca.recrown.islandsurvivalcraft.caching.Cache;
import ca.recrown.islandsurvivalcraft.datatypes.Point2;
import ca.recrown.islandsurvivalcraft.world.BiomeSelector;
import ca.recrown.islandsurvivalcraft.world.IslandWorldMapper;
import ca.recrown.islandsurvivalcraft.world.shaders.WorldHeightShader;

public class IslandWorldChunkGenerator extends ChunkGenerator implements Listener {
    private final Cache<Point2, Double> blockValueCache = new Cache<>(102400);
    private final Cache<Point2, Biome[]> biomeCache = new Cache<>(102400);
    private final Cache<Point2, Boolean> chunkExistenceCache = new Cache<>(16384);
    private final BiomeSelector biomeSelector = new BiomeSelector();
    private final ExecutorService executor = Utilities.ISC_EXECUTOR;
    private volatile World currentWorld;

    public IslandWorldChunkGenerator() {
        biomeSelector.initialize();
    }

    @Override
    public ChunkData generateChunkData(World world, Random random, int chunkX, int chunkZ, BiomeGrid biomeGrid) {
        this.currentWorld = world;
        IslandWorldMapper mapper = new IslandWorldMapper(random, blockValueCache);
        TemperatureMapGenerator temperatureMapGenerator = new TemperatureMapGenerator(world.getSeed());
        WorldHeightShader heightShader = new WorldHeightShader(world.getSeed(), mapper, world.getSeaLevel(), world.getMaxHeight(), 3);
        BiomeGenerator biomeGenerator = new UniBiomeIslandGenerator();

        int maxHeight = world.getMaxHeight();
        LinkedList<Future<Boolean>> tasks = new LinkedList<>();
        ChunkData chunkData = createChunkData(world);
        Future<Boolean> preLoader = executor.submit(() -> {
            for (int x = 0; x < Utilities.CHUNK_SIZE; x++) {
                for (int z = 0; z < Utilities.CHUNK_SIZE; z++) {
                    if (Thread.currentThread().isInterrupted()) return false;
                    mapper.getWorldBlockValue(Utilities.CHUNK_SIZE * chunkX + x, Utilities.CHUNK_SIZE * chunkZ + z);
                }
            }
            return true;
        });

        Biome[][] biomes = new Biome[Utilities.CHUNK_SIZE][Utilities.CHUNK_SIZE];
        for (int x = 0; x < Utilities.CHUNK_SIZE; x++) {
            for (int z = 0; z < Utilities.CHUNK_SIZE; z++) {
                final int desX = x;
                final int desZ = z;
                if (biomes[desX][desZ] == null) {
                    biomeGenerator.generateBiomeColumn(biomes, world, chunkX, chunkZ, desX, desZ, mapper, biomeSelector,
                            temperatureMapGenerator, biomeCache, chunkExistenceCache);
                }
                
                if (biomes[desX][desZ] == null) throw new IllegalStateException("Biome was null.");
                tasks.add(executor.submit(() -> {
                    for (int y = 0; y < maxHeight; y++) {
                        biomeGrid.setBiome(desX, y, desZ, biomes[desX][desZ]);
                    }
                    return true;
                }));

                final int worldX = Utilities.CHUNK_SIZE * chunkX + desX;
                final int worldZ = Utilities.CHUNK_SIZE * chunkZ + desZ;
                int height = heightShader.getAltitude(worldX, worldZ, biomes[desX][desZ]);
                chunkData.setRegion(desX, 1, desZ, desX + 1, height, desZ + 1, Material.DIAMOND_BLOCK);
            }
        }
        chunkData.setRegion(0, 0, 0, 16, 1, 16, Material.BEDROCK);

        preLoader.cancel(true);
        try {
            while (!tasks.isEmpty()) {
                tasks.poll().get();
            }
        } catch (ExecutionException | InterruptedException e) {
            e.printStackTrace();
            throw new IllegalStateException(e.getCause().getMessage());
        }
        return chunkData;
    }
    
    @Override
    public boolean canSpawn(World world, int x, int z) {
        return super.canSpawn(world, x, z);
    }

    @Override
    public boolean isParallelCapable() {
        return true;
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onChunkLoaded(ChunkLoadEvent event) {
        if (event.isNewChunk()) {
            if (currentWorld == event.getWorld()) {
                Chunk chunk = event.getChunk();
                
                chunkExistenceCache.set(new Point2(chunk.getX(), chunk.getZ()), true);
            }
        }
    }
}